#!/usr/bin/env python3

# Copyright 2022 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

import os

from airflow.kubernetes.secret import Secret
from airflow.models import Variable
from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import (
    KubernetesPodOperator
)
from airflow.operators.docker_operator import DockerOperator
from docker.types import Mount
from kubernetes.client.models import V1KeyToPath


class VoxrowOperator:
    '''
    >>> import os
    >>> from voxrow.airflow import VoxrowOperator
    >>> SECRET = 'put-k8s-secret-name-here'  # For KubernetesPodOperator
    >>> task_name = VoxrowOperator(
    ...     env={
    ...         SECRET: [
    ...             'DB_DRIVER',
    ...             'DB_HOST'
    ...         ]
    ...     },
    ...     mounts_volumes=[{
    ...         'secret_name': SECRET,
    ...         'items': [
    ...             {
    ...                 'key': 'creds.json',
    ...                 'path': 'creds.json',
    ...                 'read_only': True
    ...             }
    ...         ]
    ...         'mount_path': '/tmp/',
    ...         # For DockerOperator, use case:
    ...         # https://gitlab.com/pipinfitriadi/airflow-local.git
    ...         'source_path': os.path.join(
    ...             os.getenv('APP_DIR', ''),
    ...             '../..'
    ...         ),
    ...     }],
    ...     # For DockerOperator, use case:
    ...     # https://gitlab.com/pipinfitriadi/airflow-local.git
    ...     use_kubernetes=Variable.get('use_kubernetes', 'true', True),
    ...     image='alpine',
    ...     task_id='tes',
    ...     entrypoint=['ls'],
    ...     command=['-lah', '/tmp/']
    ...     # For KubernetesPodOperator
    ...     namespace='put-k8s-airflow-namespace-here',
    ...     image_pull_secrets='put-k8s-registry-secrets-here-only-if-needed',
    ... )
    >>>
    '''

    DOCKER_TEMPLATE_FIELDS: set = {
        'image',
        'api_version',
        'command',
        'container_name',
        'cpus',
        'docker_url',
        'environment',
        'private_environment',
        'env_file',
        'force_pull',
        'mem_limit',
        'host_tmp_dir',
        'network_mode',
        'tls_ca_cert',
        'tls_client_cert',
        'tls_client_key',
        'tls_hostname',
        'tls_ssl_version',
        'mount_tmp_dir',
        'tmp_dir',
        'user',
        'mounts',
        'entrypoint',
        'working_dir',
        'xcom_all',
        'docker_conn_id',
        'dns',
        'dns_search',
        'auto_remove',
        'shm_size',
        'tty',
        'hostname',
        'privileged',
        'cap_add',
        'extra_hosts',
        'retrieve_output',
        'retrieve_output_path',
        'timeout',
        'device_requests',
        'log_opts_max_size',
        'log_opts_max_file',
        'ipc_mode'
    }

    K8S_TEMPLATE_FIELDS: set = {
        'kubernetes_conn_id',
        'namespace',
        'image',
        'name',
        'random_name_suffix',
        'cmds',
        'arguments',
        'ports',
        'volume_mounts',
        'volumes',
        'env_vars',
        'env_from',
        'secrets',
        'in_cluster',
        'cluster_context',
        'labels',
        'reattach_on_restart',
        'startup_timeout_seconds',
        'get_logs',
        'image_pull_policy',
        'annotations',
        'container_resources',
        'affinity',
        'config_file',
        'node_selector',
        'image_pull_secrets',
        'service_account_name',
        'is_delete_operator_pod',
        'hostnetwork',
        'tolerations',
        'security_context',
        'container_security_context',
        'dnspolicy',
        'schedulername',
        'full_pod_spec',
        'init_containers',
        'log_events_on_failure',
        'do_xcom_push',
        'pod_template_file',
        'priority_class_name',
        'pod_runtime_info_envs',
        'termination_grace_period',
        'configmaps'
    }

    def __new__(
        cls,
        env: dict = {},
        mounts_volumes: list = [],
        use_kubernetes: bool = True,
        **kwargs
    ):
        task_id = kwargs['task_id']

        if use_kubernetes:
            kwargs['cmds'] = kwargs.get(
                'cmds', kwargs.pop('entrypoint', [])
            )
            kwargs['arguments'] = kwargs.get(
                'arguments', kwargs.pop('command', [])
            )

            for key in cls.DOCKER_TEMPLATE_FIELDS & kwargs.keys():
                if key != 'image':
                    kwargs.pop(key)

            kwargs['secrets'] = [
                *(
                    Secret('env', env_var, secret_name, env_var)
                    for secret_name, env_vars in env.items()
                    for env_var in env_vars
                ),
                *[
                    Secret(
                        'volume',
                        volume['mount_path'],
                        volume['secret_name'],
                        None,
                        # V1KeyToPath
                        # https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1KeyToPath.md
                        [
                            V1KeyToPath(
                                **{
                                    key: value
                                    for key, value in item.items()
                                    if key in ['key', 'mode', 'path']
                                }
                            )
                            for item in volume['items']
                        ]
                    )
                    for volume in mounts_volumes
                ],
                *kwargs.get('secrets', [])
            ]
            kwargs['name'] = kwargs.get('name', task_id)
            kwargs['labels'] = kwargs.get('labels', {'pod-label': task_id})
            kwargs['image_pull_policy'] = kwargs.get(
                'image_pull_policy', 'Always'
            )
            kwargs['is_delete_operator_pod'] = kwargs.get(
                'is_delete_operator_pod', True
            )
            kwargs['in_cluster'] = kwargs.get('in_cluster', True)
            kwargs['get_logs'] = kwargs.get('get_logs', True)

            operator = KubernetesPodOperator(**kwargs)
        else:
            kwargs['entrypoint'] = kwargs.get(
                'entrypoint', kwargs.pop('cmds', [])
            )
            kwargs['command'] = kwargs.get(
                'command', kwargs.pop('arguments', [])
            )

            for key in cls.K8S_TEMPLATE_FIELDS & kwargs.keys():
                if key != 'image':
                    kwargs.pop(key)

            kwargs['private_environment'] = {
                **{
                    key: Variable.get(
                        key.lower()
                    )
                    for value in env.values()
                    for key in value
                },
                **kwargs.get('private_environment', {})
            }
            kwargs['mounts'] = [
                *[
                    Mount(
                        **{
                            key: os.path.join(
                                volume[value], item['path']
                            )
                            for key, value in {
                                'source': 'source_path',
                                'target': 'mount_path'
                            }.items()
                        },
                        type='bind',
                        read_only=item['read_only']
                    )
                    for volume in mounts_volumes
                    for item in volume['items']
                ],
                *kwargs.get('mounts', [])
            ]
            kwargs['container_name'] = kwargs.get('container_name', task_id)
            kwargs['mount_tmp_dir'] = kwargs.get('mount_tmp_dir', False)
            kwargs['auto_remove'] = kwargs.get('auto_remove', True)
            kwargs['docker_url'] = kwargs.get(
                'docker_url', 'tcp://docker-proxy:2375'
            )

            operator = DockerOperator(**kwargs)

        return operator



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

import __main__ as main
from datetime import date, datetime
from json import (
    dumps,
    JSONDecoder as _JSONDecoder,
    JSONEncoder as _JSONEncoder,
    loads
)
from json.decoder import JSONDecodeError, WHITESPACE
from pathlib import Path
import re
from statistics import mean
from sys import exc_info
from traceback import print_exception

STRING_DATE_FORMAT = '%Y-%m-%d'
STRING_DATETIME_FORMAT = f'{ STRING_DATE_FORMAT }T%H:%M:%S'


def display_time(seconds, granularity=2, sort_name=False):
    def text_result(value, name, sort_name):
        if sort_name:
            name = name[0]
        elif value in range(2):
            name = name.rstrip('s')

        return f"{ value }{ '' if sort_name else ' ' }{ name }"

    # Python function to convert seconds into minutes, hours, and days
    # https://stackoverflow.com/questions/4048651/python-function-to-convert-seconds-into-minutes-hours-and-days
    def result(seconds):
        check_granularity = 0
        SECOND = 1
        MINUTE = SECOND * 60
        HOUR = MINUTE * 60
        DAY = HOUR * 24
        WEEK = DAY * 7

        for name, count in {
            'weeks': WEEK,
            'days': DAY,
            'hours': HOUR,
            'minutes': MINUTE,
            'seconds': SECOND,
        }.items():
            value = seconds // count

            if value:
                yield text_result(value, name, sort_name)
                check_granularity += 1

                if check_granularity >= granularity:
                    break
                else:
                    seconds -= value * count
        else:
            if not check_granularity:
                yield text_result(value, name, sort_name)

    return (' ' if sort_name else ', ').join(
        list(
            result(seconds)
        )
    )


class Log:
    def __init__(self, file_name=None):
        self.__PROCESS_TIME = []
        self.__START_TIME = datetime.now()
        self.__LAST_TIME = self.__START_TIME
        self.__FILE_NAME = file_name

        if not self.__FILE_NAME and hasattr(main, '__file__'):
            # Get name of current script in Python
            # https://stackoverflow.com/questions/4152963/get-name-of-current-script-in-python
            self.__FILE_NAME = f'{ blueprint_name(main.__file__) }.log'

        if self.__FILE_NAME:
            with open(self.__FILE_NAME, 'w'):
                pass

    def __call__(self, *data, sep=' | ', end='\n'):
        last_time = datetime.now()
        self.__PROCESS_TIME.append(
            (last_time - self.__LAST_TIME).seconds
        )
        self.__LAST_TIME = last_time
        total_time = (self.__LAST_TIME - self.__START_TIME).seconds
        average_time = round(
            mean(self.__PROCESS_TIME)
        )

        for d in data:
            if isinstance(d, Exception):
                is_error = True
                error = d
                break
        else:
            is_error = False

        if not is_error:
            data = [
                self.__LAST_TIME.strftime(STRING_DATETIME_FORMAT),
                f'TOTAL: { display_time(total_time, 3, True) }',
                f'AVG: { display_time(average_time, 3, True) }',
                *data
            ]
            print(
                *data,
                sep=sep,
                end=end
            )

        if self.__FILE_NAME:
            # Convert bytes to a string
            # https://stackoverflow.com/questions/606191/convert-bytes-to-a-string
            # How to seek and append to a binary file in python
            # https://stackoverflow.com/questions/4388201/how-to-seek-and-append-to-a-binary-file-in-python
            # Remove very last character in file
            # https://stackoverflow.com/questions/18857352/remove-very-last-character-in-file
            # Writing to a File with Python's print() Function
            # https://stackabuse.com/writing-to-a-file-with-pythons-print-function/
            with open(self.__FILE_NAME, 'a') as log_file:
                if is_error:
                    # Python: How to write error in the console in txt file?
                    # https://stackoverflow.com/questions/55169364/python-how-to-write-error-in-the-console-in-txt-file
                    # How to print the full traceback without halting the
                    # program
                    # https://stackoverflow.com/questions/3702675/how-to-print-the-full-traceback-without-halting-the-program
                    print_exception(*exc_info(), file=log_file)
                else:
                    print(*data, sep=sep, end=end, file=log_file)

        if is_error:
            raise error


def blueprint_name(file):
    # How to get the filename without the extension from a path in Python?
    # https://stackoverflow.com/questions/678236/how-to-get-the-filename-without-the-extension-from-a-path-in-python
    path = Path(file)
    name = (path).resolve().stem

    if name == '__init__':
        name = path.parent.resolve().stem

    return name


# How to convert string int JSON into real int with json.loads
# https://stackoverflow.com/questions/45068797/how-to-convert-string-int-json-into-real-int-with-json-loads
# How to convert to a Python datetime object with JSON.loads?
# https://stackoverflow.com/questions/8793448/how-to-convert-to-a-python-datetime-object-with-json-loads
def deserialize(object):
    '''
    >>> deserialize('2020-04-28'), deserialize('2020-04-28T15:00:46')
    '''

    if isinstance(object, str):
        try:
            object = loads(object)
        except JSONDecodeError:
            pass

        if isinstance(object, str):
            for regex, str_time_format in [
                [
                    r'^\d{4}-\d{2}-\d{2}$',
                    STRING_DATE_FORMAT
                ],
                [
                    r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$',
                    STRING_DATETIME_FORMAT
                ]
            ]:
                if re.match(regex, object):
                    is_time = False

                    try:
                        object = datetime.strptime(object, str_time_format)

                        is_time = True
                    except Exception:
                        pass

                    if is_time:
                        break
        else:
            object = deserialize(object)
    elif isinstance(object, dict):
        object = {
            key: deserialize(value)
            for key, value in object.items()
        }
    elif isinstance(object, list):
        object = [
            deserialize(value)
            for value in object
        ]

    return object


class JSONDecoder(_JSONDecoder):
    def decode(self, s, _w=WHITESPACE.match):
        return deserialize(
            super().decode(s, _w)
        )


class JSONEncoder(_JSONEncoder):
    '''
    >>> from json import dumps
    >>> dumps(object, cls=JSONEncoder)
    '''

    # Custom JSON encoder for Flask
    # https://gist.github.com/claraj/3b2b95a62c5ba6860c03b5c737c214ab
    # Pass user built json encoder into Flask's jsonify
    # https://stackoverflow.com/questions/44146087/pass-user-built-json-encoder-into-flasks-jsonify/44158611
    def default(self, obj):
        return (
            super().default(obj)
            if isinstance(obj, str)
            else serialize(obj)
        )


def serialize(object):
    '''
    >>> from json import dumps
    >>> dumps(object, default=serialize)
    '''

    # How to serialize a datetime object as JSON using Python?
    # https://code-maven.com/serialize-datetime-object-as-json-in-python
    if isinstance(object, datetime):
        object = object.strftime(STRING_DATETIME_FORMAT)
    elif isinstance(object, date):
        object = object.strftime(STRING_DATE_FORMAT)
    elif isinstance(object, set):
        object = list(object)

    return object


def json_serializer(object):
    return dumps(object, cls=JSONEncoder)


def json_deserializer(object):
    return loads(object, cls=JSONDecoder)



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from types import GeneratorType

from flask import (
    Flask as _Flask,
    jsonify,
    render_template,
    Response,
    request,
    stream_with_context,
    url_for
)
from jinja2.exceptions import TemplateNotFound
from werkzeug.datastructures import Headers
from werkzeug.exceptions import default_exceptions
from werkzeug.http import HTTP_STATUS_CODES
from werkzeug.routing import BuildError, RequestRedirect
from werkzeug.wrappers import Response as ResponseBase

try:
    from ...config import WEB_TITLE
except Exception:
    WEB_TITLE = 'VOXROW'


class Flask(_Flask):
    # Customizing the Flask Response Class
    # https://blog.miguelgrinberg.com/post/customizing-the-flask-response-class

    def __init__(self, *args, **kwargs):
        '''
        Dipergunakan untuk mengabaikan custome func make_response.
        exclude_endpoint: str
        '''

        self.exclude_endpoint = kwargs.pop('exclude_endpoint', None)
        super().__init__(*args, **kwargs)

    def make_response(self, rv):
        '''
        Pilihan value yang dapat dipergunakan pada function:
        - rv = response
        - rv = (response, status, headers, template, template_mimetype)
        - rv = (response, status, headers, template)
        - rv = (response, headers, template, template_mimetype)
        - rv = (response, status, template, template_mimetype)
        - rv = (response, template, template_mimetype)
        - rv = (response, headers, template)
        - rv = (response, status, template)
        - rv = (response, template)
        - rv = (response, headers)
        - rv = (response, status)

        Tipe data:
        - template: str = None
        - template_mimetype: str = 'text/html'
        - template_mimetype: bool = True
            Apabila nilainya bool True, maka tidak akan dilakukan pengecekan
            Header Accept Request.
        '''

        status = headers = template = None
        template_mimetype = 'text/html'

        if isinstance(rv, tuple):
            header_type = (Headers, dict, tuple, list)
            len_rv = len(rv)

            if len_rv == 5:
                rv, status, headers, template, template_mimetype = rv
            elif len_rv == 4:
                rv, status, headers, template = rv
            elif len_rv == 3:
                if isinstance(rv[2], str):
                    if isinstance(rv[1], str):
                        rv, template, template_mimetype = rv
                    elif isinstance(rv[1], header_type):
                        rv, headers, template = rv
                    else:
                        rv, status, template = rv
                else:
                    rv, status, headers = rv
            elif len_rv == 2:
                if isinstance(rv[1], str):
                    rv, template = rv
                elif isinstance(rv[1], header_type):
                    rv, headers = rv
                else:
                    rv, status = rv
            else:
                raise TypeError(
                    'The view function did not return a valid'
                    ' response. The return type must be a string, dict, tuple'
                    ' (with max length no more than 5), Response instance, or'
                    ' WSGI callable, but it was a'
                    f' {rv.__class__.__name__}.'
                )

        if (
            not isinstance(rv, self.response_class)
            and not isinstance(rv, ResponseBase)
            and not isinstance(rv, RequestRedirect)
        ):
            if status and status >= 400:
                if isinstance(rv, dict):
                    message = rv.get('error_message')
                    name = rv.get('error_name')
                    parameter = rv.get('error_parameter')
                else:
                    message = default_exceptions[status].description
                    # Flask - How to create custom abort() code?
                    # https://stackoverflow.com/questions/12285903/flask-how-to-create-custom-abort-code/38648607#38648607
                    name = HTTP_STATUS_CODES[status]
                    parameter = {}

                message = str(message) if message else ''

                try:
                    message = (
                        f'{ message } '
                        f'To make { WEB_TITLE } works, please see the'
                        ' Doc in here:'
                        f' { url_for("doc.show", _external=True) }'
                    )
                except BuildError:
                    pass

                data = {
                    'data': None,
                    'error': {
                        'code': status,
                        'message': message,
                        'name': name,
                        'parameter': parameter if parameter else {}
                    }
                }
            elif isinstance(rv, GeneratorType):
                # How to check if an object is a generator object in python
                # https://stackoverflow.com/questions/6416538/how-to-check-if-an-object-is-a-generator-object-in-python
                def generator(rv):
                    yield '{"data": '
                    yield from rv
                    yield ', "error": null}'

                data = generator(rv)
            else:
                data = {
                    'data': rv,
                    'error': None
                }

            if (
                (endpoint := request.endpoint) != self.exclude_endpoint
                or (
                    endpoint == self.exclude_endpoint
                    and (
                        endpoint is None
                        or url_for(endpoint) != request.path
                    )
                )
            ):
                if (
                    isinstance(rv, GeneratorType)
                    and (
                        status is None
                        or (
                            status
                            and status < 400
                        )
                    )
                ):
                    # Python flask.stream_with_context() Examples
                    # https://www.programcreek.com/python/example/58918/flask.stream_with_context
                    rv = Response(
                        stream_with_context(data),
                        mimetype='application/json'
                    )
                elif (
                    template
                    and (
                        template_mimetype is True
                        or request.accept_mimetypes.best == template_mimetype
                    )
                ):
                    try:
                        rv = render_template(
                            template,
                            data=data
                        )
                    except TemplateNotFound:
                        rv = jsonify(data)
                else:
                    rv = jsonify(data)
        elif isinstance(rv, ResponseBase):
            try:
                if (
                    '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n'
                    "<title>Redirecting...</title>\n"
                    "<h1>Redirecting...</h1>\n"
                    "<p>You should be redirected automatically to target URL: "
                    '<a href="'
                ) in rv.data.decode('utf-8'):
                    rv.data = render_template('redirecting.html').encode()
            except RuntimeError:
                pass
            except TemplateNotFound:
                pass

        return super().make_response(
            (rv, status, headers)
        )



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask import Blueprint
from sqlalchemy.schema import CreateSchema, DropSchema

from ... import blueprint_name
from ...extension.model import db, User
from ...modify.flask_sqlalchemy import DB_SCHEMA

app = Blueprint(
    (
        app_name := blueprint_name(__file__)
    ),
    __name__,
    cli_group=app_name
)


def database_setup():
    with db.engine.connect() as conn:
        # SQLAlchemy: “create schema if not exists”
        # https://stackoverflow.com/questions/50927740/sqlalchemy-create-schema-if-not-exists
        if not conn.dialect.has_schema(conn, DB_SCHEMA):
            # Getting SQLAlchemy to issue CREATE DB SCHEMA on create_all
            # https://stackoverflow.com/questions/13677781/getting-sqlalchemy-to-issue-create-schema-on-create-all
            db.engine.execute(
                CreateSchema(DB_SCHEMA)
            )

    db.create_all()

    if not User.read():
        User.create(
            User(
                username='admin',
                password='admin1234',
                is_admin=True
            )
        )


@app.before_app_first_request
def before_app_first_request():
    database_setup()


@app.cli.command('create')
def create_database():
    # Command Line Interface
    # https://flask.palletsprojects.com/en/1.1.x/cli/
    database_setup()


@app.cli.command('drop')
def drop_database():
    db.drop_all()

    with db.engine.connect() as conn:
        if conn.dialect.has_schema(conn, DB_SCHEMA):
            db.engine.execute(
                DropSchema(DB_SCHEMA)
            )



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask_sqlalchemy.model import Model as _Model
from sqlalchemy import inspect
from sqlalchemy.sql.elements import BinaryExpression, BooleanClauseList


class Model(_Model):
    db = None

    def to_dict(self):
        # Convert sqlalchemy row object to python dict
        # https://stackoverflow.com/questions/1958219/convert-sqlalchemy-row-object-to-python-dict
        return {
            parameter.key: getattr(self, parameter.key)
            for parameter in inspect(self).mapper.column_attrs
        }

    def __iter__(self):
        # Python dictionary from an object's fields
        # https://stackoverflow.com/questions/61517/python-dictionary-from-an-objects-fields
        for key, value in self.to_dict().items():
            yield key, value

    @classmethod
    def create_session(cls, options=None):
        # Commit changes for only one SQLAlchemy model instance when multiple
        # have changed
        # https://stackoverflow.com/questions/37191141/commit-changes-for-only-one-sqlalchemy-model-instance-when-multiple-have-changed
        # Python sqlalchemy.orm.sessionmaker() Examples
        # https://www.programcreek.com/python/example/51135/sqlalchemy.orm.sessionmaker
        return cls.db.create_session(
            options if isinstance(options, dict)
            else {}
        )()

    @classmethod
    def commit(cls, session, only_read=False):
        try:
            if not only_read:
                session.commit()
        # KeyboardInterrupt and SystemExit should not be wrapped by
        # sqlalchemy #689
        # https://github.com/sqlalchemy/sqlalchemy/issues/689
        # Avoiding accidentally catching KeyboardInterrupt and
        # SystemExit in Python 2.4
        # https://stackoverflow.com/questions/2669750/avoiding-accidentally-catching-keyboardinterrupt-and-systemexit-in-python-2-4
        # Except block handles 'BaseException'
        # https://lgtm.com/rules/6780080/
        # Catch multiple exceptions in one line (except block)
        # https://stackoverflow.com/questions/6470428/catch-multiple-exceptions-in-one-line-except-block
        # How to get the process ID to kill a nohup process?
        # https://stackoverflow.com/questions/17385794/how-to-get-the-process-id-to-kill-a-nohup-process
        # Fixing “Lock wait timeout exceeded; try restarting
        # transaction” for a 'stuck" Mysql table?
        # https://stackoverflow.com/questions/2766785/fixing-lock-wait-timeout-exceeded-try-restarting-transaction-for-a-stuck-my/10315184
        except (KeyboardInterrupt, SystemExit, Exception):
            session.rollback()
            raise
        finally:
            session.close()

    @classmethod
    def create(cls, row):
        if not isinstance(row, cls):
            raise TypeError(
                f'It should be { cls.__name__ } instance!'
            )

        (
            session := cls.create_session()
        ).add(row)
        cls.commit(session)

    @classmethod
    def read(cls, criterion=None):
        session = cls.create_session()

        if (
            isinstance(criterion, BinaryExpression)
            or isinstance(criterion, BooleanClauseList)
            or isinstance(criterion, bool)
        ):
            data = session.query(cls).filter(criterion).all()
        elif criterion is None:
            data = session.query(cls).all()
        else:
            data = session.query(cls).get(criterion)

        cls.commit(session, True)
        return data

    @classmethod
    def update(
        cls,
        criterion,
        values,
        synchronize_session='evaluate',
        update_args=None
    ):
        data = (
            session := cls.create_session()
        ).query(cls).filter(criterion).update(
            values,
            synchronize_session,
            update_args
        )
        cls.commit(session)
        return data

    @classmethod
    def delete(cls, ident=None):
        session = cls.create_session()
        result = None

        if ident and (
            data := cls.read(ident)
        ):
            session.delete(data)
        elif ident is None:
            result = session.query(cls).delete()

        cls.commit(session)
        return result



#!/usr/bin/env python3

# Copyright 2022 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from collections.abc import Iterable
from datetime import date, datetime
from json import dumps
import logging
from os import getenv, getcwd
from os.path import isfile, join as path_join
import re
from time import sleep
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

from jinja2 import Environment, FileSystemLoader
from jinja2.meta import find_referenced_templates, find_undeclared_variables
from sshtunnel import SSHTunnelForwarder
from sqlalchemy import bindparam, create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.exc import OperationalError, StatementError
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm.query import Query as _Query
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.sql import text
from sqlalchemy.sql.expression import null
from sqlalchemy.types import (
    Boolean,
    Date,
    DateTime,
    Float,
    Integer,
    JSON,
    NullType,
    String
)

from . import deserialize, json_serializer, serialize


def database_uri(
    db_driver='postgresql',
    db_host='localhost',
    db_port=5432,
    db_name='postgres',
    db_user='',
    db_pass='',
    use_charset_utf8=False
):
    if db_driver == 'mysql' and db_port == 5432:
        db_port = 3306

    uri = (
        (
            f'{ db_driver }://{ db_user }:{ db_pass }@'
            f'{ db_host }:{ db_port }/{ db_name }'
        )
        if db_driver else ''
    )

    # flask sqlalchemy mysql encoding problems
    # https://stackoverflow.com/questions/26577334/flask-sqlalchemy-mysql-encoding-problems
    # SQLAlchemy + MySQL + UTF-8 support - how?
    # https://groups.google.com/forum/#!topic/pylons-discuss/ol2m46kiSYA
    return (
        uri + '?charset=utf8'
        if db_driver == 'mysql'
        and use_charset_utf8
        else uri
    )


def database_uri_from_env(
    db_driver_env='DB_DRIVER',
    db_host_env='DB_HOST',
    db_port_env='DB_PORT',
    db_name_env='DB_NAME',
    db_user_env='DB_USER',
    db_pass_env='DB_PASS',
    use_charset_utf8_env='USE_CHARSET_UTF8',
):
    return database_uri(
        getenv(db_driver_env, 'postgresql'),
        getenv(db_host_env, 'localhost'),
        int(
            getenv(db_port_env, '5432')
        ),
        getenv(db_name_env, 'postgres'),
        getenv(db_user_env, ''),
        getenv(db_pass_env, ''),
        getenv(use_charset_utf8_env, 'FALSE').upper() in ['TRUE', '1']
    )


# Retry failed sqlalchemy queries
# https://stackoverflow.com/questions/53287215/retry-failed-sqlalchemy-queries/60614707#60614707
class RetryingQuery(_Query):
    __max_retry_count__ = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __iter__(self):
        attempts = 0

        while True:
            attempts += 1

            try:
                return super().__iter__()
            except OperationalError as ex:
                if 'server closed the connection unexpectedly' not in str(ex):
                    raise

                if attempts <= self.__max_retry_count__:
                    sleep_for = 2 ** (attempts - 1)
                    logging.error(
                        ' [!] Database connection error: retrying Strategy => '
                        f'sleeping for {sleep_for}s and will retry '
                        f'(attempt #{attempts} of {self.__max_retry_count__})'
                        f'\nDetailed query impacted: {ex}'
                    )
                    sleep(sleep_for)
                    continue
                else:
                    raise
            except StatementError as ex:
                if (
                    'reconnect until invalid transaction is rolled back'
                    not in str(ex)
                ):
                    raise
                self.session.rollback()


class Query:
    def __init__(self, engine, use_charset_utf8=False, template_dir=None):
        '''
        engine:
        - obj: SQLAlchemy's engine instance.
        - str: Build SQLAlchemy's engine from string database uri.
        - dict: Build SQLAlchemy's engine from database_uri's func kwargs and
        SSHTunnelForwarder's func kwarg.

        use_charset_utf8: bool
        - Use use_charset_utf8=True for mysql driver if needed.

        template_dir: str
        - Use for set specify jinja2 template directory, os.getcwd()
        value is used as default.
        '''

        self.__template_dir = template_dir if template_dir else getcwd()
        self.__env = Environment(
            loader=FileSystemLoader(self.__template_dir)
        )
        self.__ssh_param = {}
        self.__engine = engine
        self.__use_charset_utf8 = use_charset_utf8

        if isinstance(self.__engine, dict):
            self.__db_host = None
            self.__db_port = None
            database_uri_param = {
                'db_driver',
                'db_host',
                'db_port',
                'db_name',
                'db_user',
                'db_pass',
                'use_charset_utf8'
            }
            self.__ssh_param = {
                key: value
                for key, value in self.__engine.items()
                if key not in database_uri_param
            }
            self.__engine = {
                key: value
                for key, value in self.__engine.items()
                if key in database_uri_param
            }

            if self.__use_charset_utf8:
                self.__engine['use_charset_utf8'] = self.__use_charset_utf8
            elif use_charset_utf8 := self.__engine.get('use_charset_utf8'):
                self.__use_charset_utf8 = use_charset_utf8

    def __call__(self, string, **kwargs):
        '''
        string: str
        - Use for put raw query sql or sql file path. It is support jinja2
        templating.

        kwargs: dict
        - Put sql parameter and or jinja2 variable in here.
        - Optional parameter can be use for best query result:
            1. generator_mode: bool
                - Use generator_mode=True if we want result as generator.
            2. json_mode: bool
                - Use json_mode=True if we want result as string dumps json.
            3. debug_mode: bool
                - Use debug_mode=False if you don't want to see error
                information.
            4. fetch_size: int
                - Use for set how many rows use on every fetch.
                - Default value is 100,000.
            5. stream_results: bool
                - Use stream_results=False for update/insert/delete query.
        '''

        if self.__ssh_param:
            ssh_param = self.__ssh_param
            self.__ssh_param = {}

            while True:
                with SSHTunnelForwarder(**ssh_param) as tunnel:
                    # Python SSHTunnel w/ Paramiko - CLI works, but
                    # not in script
                    # https://stackoverflow.com/questions/39945269/python-sshtunnel-w-paramiko-cli-works-but-not-in-script
                    # Menjalankan query dengan mode SSH
                    # SSHTunnelForwarder's func kwargs
                    # https://sshtunnel.readthedocs.io/en/latest/#api
                    # Setup a SSH Tunnel With the Sshtunnel Module in Python
                    # https://blog.ruanbekker.com/blog/2018/04/23/setup-a-ssh-tunnel-with-the-sshtunnel-module-in-python/
                    # SSHTunnelForwarder.daemon_forward_servers is not
                    # respected:
                    # https://github.com/pahaz/sshtunnel/issues/102
                    # tunnel without clause:
                    # tunnel.daemon_forward_servers = True
                    # tunnel.daemon_transport = True
                    # tunnel.start()
                    # tunnel.stop()
                    sleep(1)

                    self.__db_host = tunnel.local_bind_host
                    self.__db_port = tunnel.local_bind_port

                    try:
                        result = self.__call__(string, **kwargs)
                        break
                    except OperationalError as e:
                        if (
                            e.args
                            and e.orig
                            and e.orig.args
                            and (
                                # Error MySQL: Lost Connection
                                'mysql' in e.args[0].lower()
                                and e.orig.args[0] == 2013
                            )
                        ):
                            continue
                        else:
                            raise
                    except (KeyboardInterrupt, SystemExit):
                        break
                    except Exception:
                        raise
                    finally:
                        self.__db_host = None
                        self.__db_port = None

            self.__ssh_param = ssh_param
        else:
            # Issue with a python function returning a generator or a normal
            # object
            # https://stackoverflow.com/questions/25313283/issue-with-a-python-function-returning-a-generator-or-a-normal-object
            generator_mode = kwargs.pop('generator_mode', None)

            result = self.__query(string, **kwargs)

            if not generator_mode:
                result = (
                    ''.join(result)
                    if kwargs.get('json_mode') is True
                    else list(result)
                )

        return result

    def __compile(self, text_clause):
        '''
        Function to compile and bind parameter sql.

        text_clause: sqlalchemy.sql.elements.TextClause
        '''

        # SQLAlchemy: print the actual query
        # https://stackoverflow.com/questions/5631078/sqlalchemy-print-the-actual-query
        class StringLiteral(String):
            '''
            Teach SA how to literalize various things.
            '''

            def literal_processor(self, dialect):
                super_processor = super(
                    StringLiteral, self
                ).literal_processor(dialect)

                def process(value):
                    if (
                        isinstance(
                            value := serialize(value),
                            Iterable
                        )
                        and not isinstance(value, str)
                    ):
                        value = dumps(value)

                    if isinstance(
                        result := super_processor(value),
                        bytes
                    ):
                        result = result.decode(dialect.encoding)

                    return result

                return process

        class LiteralDialect(DefaultDialect):
            colspecs = {
                DateTime: StringLiteral,
                Date: StringLiteral,
                JSON: StringLiteral
            }

        return str(
            text_clause.compile(
                dialect=LiteralDialect(),
                compile_kwargs={'literal_binds': True}
            )
        )

    def render(self, string, literal_binds=True, **kwargs):
        '''
        Function for render jinja2 template and bind parameter sql.

        string: str
        - Use for put raw query sql or sql file path. It is support jinja2
        templating.

        literal_binds: bool
        - Use literal_binds=False if not want to see sql bind parameter.

        kwargs: dict
        - Put sql parameter and or jinja2 variable in here.
        - It is mandatory if string have sql parameter and literal_binds=True.
        '''

        # SQLAlchemy: print the actual query
        # https://stackoverflow.com/questions/5631078/sqlalchemy-print-the-actual-query

        def __template_source(template_file):
            '''
            Function for returning jinja2 template source.

            template_file: str
            - Template file path.
            '''

            return self.__env.loader.get_source(self.__env, template_file)[0]

        def jinja2_keys(template_source):
            # Jinja2 load templates from separate location than working
            # directory
            # https://stackoverflow.com/questions/37968787/jinja2-load-templates-from-separate-location-than-working-directory
            # Template Designer Documentation
            # https://jinja.palletsprojects.com/en/2.11.x/templates/
            # How to get list of all variables in jinja 2 templates
            # https://stackoverflow.com/questions/8260490/how-to-get-list-of-all-variables-in-jinja-2-templates
            # The Meta API
            # https://jinja.palletsprojects.com/en/2.11.x/api/
            (
                keys := find_undeclared_variables(
                    parsed_content := self.__env.parse(template_source)
                )
            ).update({
                key
                for ref_template in find_referenced_templates(
                    parsed_content
                )
                for key in jinja2_keys(
                    __template_source(ref_template)
                )
            })
            return keys

        kwargs = {
            key: value
            for key, value in kwargs.items()
            if key not in [
                'json_mode',
                'debug_mode',
                'fetch_size',
                'stream_results'
            ]
        }

        if isinstance(string, str):
            if isfile(
                path_join(self.__template_dir, string)
            ):
                template = self.__env.get_template(string)
                template_source = __template_source(string)
            else:
                template = self.__env.from_string(string)
                template_source = string

            string = template.render(**kwargs)

            for key in (
                jinja2_keys(template_source).intersection(
                    kwargs.keys()
                ) - set(
                    re.findall(
                        r':(\w+)',
                        string
                    )
                )
            ):
                kwargs.pop(key, None)

        args = []

        for key in kwargs:
            param_kwargs = {
                'key': key,
                'type_': String
            }

            if (
                value := kwargs[key]
            ) is None:
                # How to insert NULL value in SQLAlchemy?
                # https://stackoverflow.com/questions/32959336/how-to-insert-null-value-in-sqlalchemy
                kwargs[key] = null()

            for parameter_type, database_column_type in [
                [type(None), NullType],
                [float, Float],
                [int, Integer],
                [bool, Boolean],
                [date, Date],
                [datetime, DateTime],
                [dict, JSON],
                [Iterable, JSON]
            ]:
                if (
                    isinstance(value, parameter_type)
                    and not isinstance(value, str)
                ):
                    param_kwargs['type_'] = database_column_type
                    break

            args.append(
                bindparam(**param_kwargs)
            )

        string = text(string).bindparams(*args, **kwargs)

        if literal_binds:
            string = self.__compile(string)
        elif literal_binds is not None:
            string = str(string)

        return string

    def __query(self, string, **kwargs):
        '''
        Func to get sql query result.

        string: str
        - Use for put raw query sql or sql file path. It is support jinja2
        templating.

        kwargs: dict
        - Put sql parameter and or jinja2 variable in here.
        - Optional parameter can be use for best query result:
            1. json_mode: bool
                - Use json_mode=True if we want result as string dumps json.
            2. debug_mode: bool
                - Use debug_mode=False if you don't want to see error
                information.
            3. fetch_size: int
                - Use for set how many rows use on every fetch.
                - Default value is 100,000.
            4. stream_results: bool
                - Use stream_results=False for update/insert/delete query.
        '''

        string = self.render(string, None, **kwargs)

        if (
            stream_results := kwargs.get('stream_results')
        ) is None:
            for s in re.findall(
                r"'[^']*'",
                _string := '\n'.join([
                    s
                    for s in re.sub(
                        r'--.*', '', str(string)
                    ).split('\n')
                    if s
                ]),
                flags=re.DOTALL
            ):
                _string = re.sub(
                    s.replace('[', r'\[').replace(']', r'\]'),
                    "''",
                    _string,
                    1
                )

            regex_sql_space = r'\s+(.*\s+)?'

            # Chapter 13 SQL Statements
            # https://dev.mysql.com/doc/refman/5.6/en/sql-statements.html
            for regex in [
                'EXPLAIN',
                'INSERT',
                'SET',
                regex_sql_space.join(['DELETE', 'FROM']),
                regex_sql_space.join(['MERGE', 'INTO', 'USING']),
                'CREATE',
                'ALTER',
                'DROP',
                regex_sql_space.join(['RENAME', 'TABLE']),
                regex_sql_space.join(['TRUNCATE', 'TABLE']),
                'SHOW',
                'DESCRIBE',
                'CALL',
                'DO',
                'HANDLER',
                'LOAD',
                'REPLACE',
                'KILL',
                'GRANT'
            ]:
                if re.findall(
                    fr'\s*{ regex }\s+.*',
                    _string,
                    flags=re.DOTALL + re.IGNORECASE
                ):
                    stream_results = False
                    break
            else:
                stream_results = True
        else:
            stream_results = False

        # Process to returning string database uri.
        if isinstance(self.__engine, dict):
            engine = self.__engine

            if all([
                self.__db_host,
                self.__db_port
            ]):
                engine.update({
                    'db_host': self.__db_host,
                    'db_port': self.__db_port
                })

            url = database_uri(**engine)
        elif isinstance(self.__engine, Engine):
            url = str(self.__engine.url)
        else:
            url = self.__engine

        # "set character set" in sqlalchemy?
        # https://groups.google.com/forum/#!topic/sqlalchemy/3kiPusCy8FM
        if (
            url.startswith('mysql')
            and 'charset=utf8' not in url
            and self.__use_charset_utf8
        ):
            # Add params to given URL in Python
            # https://stackoverflow.com/questions/2506379/add-params-to-given-url-in-python
            (
                query := dict(
                    parse_qsl(
                        (
                            url := list(
                                urlparse(url)
                            )
                        )[4]
                    )
                )
            ).update({'charset': 'utf8'})
            url[4] = urlencode(query)
            url = urlunparse(url)

        # scoped_session(sessionmaker()) or plain sessionmaker() in sqlalchemy?
        # https://stackoverflow.com/questions/6519546/scoped-sessionsessionmaker-or-plain-sessionmaker-in-sqlalchemy
        # how to fix “OperationalError: (psycopg2.OperationalError) server
        # closed the connection unexpectedly”
        # https://stackoverflow.com/questions/55457069/how-to-fix-operationalerror-psycopg2-operationalerror-server-closed-the-conn
        session = scoped_session(
            sessionmaker(
                bind=(
                    engine := create_engine(
                        url,
                        json_serializer=json_serializer,
                        pool_size=10,
                        max_overflow=2,
                        pool_recycle=300,
                        pool_pre_ping=True,
                        pool_use_lifo=True
                    )
                ).execution_options(stream_results=stream_results),
                query_cls=RetryingQuery
            )
        )()

        try:
            query_result = session.execute(string)

            # memory-efficient built-in SqlAlchemy iterator /
            # generator:
            # https://stackoverflow.com/questions/7389759/memory-efficient-built-in-sqlalchemy-iterator-generator
            # Python: Using Flask to stream chunked dynamic content to end
            # users
            # https://fabianlee.org/2019/11/18/python-using-flask-to-stream-chunked-dynamic-content-to-end-users/
            # Streaming Contents
            # https://flask.palletsprojects.com/en/1.1.x/patterns/streaming/#basic-usage
            # Streaming JSON with Flask
            # https://blog.al4.co.nz/2016/01/streaming-json-with-flask/
            if (json_mode := kwargs.get('json_mode') is True):
                yield '['

            while True:
                batch = query_result.fetchmany(
                    kwargs.get('fetch_size', 100_000)
                ) if query_result.returns_rows else None

                def to_result(row, json_mode=False):
                    data = dict(row)
                    return (
                        json_serializer(data)
                        if json_mode else deserialize(data)
                    )

                if not batch:
                    if not query_result.returns_rows:
                        yield to_result(
                            {
                                'affected_row': query_result.rowcount,
                                'query': self.__compile(string),
                                'finish_time': datetime.now()
                            },
                            json_mode
                        )

                    break

                rows = batch.__iter__()

                try:
                    prev_row = next(rows)

                    for row in rows:
                        result = to_result(prev_row, json_mode)
                        prev_row = row
                        yield result + ', ' if json_mode else result

                    yield to_result(prev_row, json_mode)
                except StopIteration:
                    pass
                # KeyboardInterrupt and SystemExit should not be wrapped by
                # sqlalchemy #689
                # https://github.com/sqlalchemy/sqlalchemy/issues/689
                # Avoiding accidentally catching KeyboardInterrupt and
                # SystemExit in Python 2.4
                # https://stackoverflow.com/questions/2669750/avoiding-accidentally-catching-keyboardinterrupt-and-systemexit-in-python-2-4
                # Except block handles 'BaseException'
                # https://lgtm.com/rules/6780080/
                # Catch multiple exceptions in one line (except block)
                # https://stackoverflow.com/questions/6470428/catch-multiple-exceptions-in-one-line-except-block
                # How to get the process ID to kill a nohup process?
                # https://stackoverflow.com/questions/17385794/how-to-get-the-process-id-to-kill-a-nohup-process
                # Fixing “Lock wait timeout exceeded; try restarting
                # transaction” for a 'stuck" Mysql table?
                # https://stackoverflow.com/questions/2766785/fixing-lock-wait-timeout-exceeded-try-restarting-transaction-for-a-stuck-my/10315184
                except (KeyboardInterrupt, SystemExit, Exception):
                    break

            if json_mode:
                yield ']'

            query_result.close()
            session.commit()
        except (KeyboardInterrupt, SystemExit, Exception):
            session.rollback()

            if kwargs.get('debug_mode') in [None, True]:
                raise
        finally:
            session.close()
            engine.dispose()



#!/usr/bin/env python3

# Copyright 2020 Pipin Fitriadi <pipinfitriadi@gmail.com>

# Licensed under the Microsoft Reference Source License (MS-RSL)

# This license governs use of the accompanying software. If you use the
# software, you accept this license. If you do not accept the license, do not
# use the software.

# 1. Definitions

# The terms "reproduce," "reproduction" and "distribution" have the same
# meaning here as under U.S. copyright law.

# "You" means the licensee of the software.

# "Your company" means the company you worked for when you downloaded the
# software.

# "Reference use" means use of the software within your company as a reference,
# in read only form, for the sole purposes of debugging your products,
# maintaining your products, or enhancing the interoperability of your
# products with the software, and specifically excludes the right to
# distribute the software outside of your company.

# "Licensed patents" means any Licensor patent claims which read directly on
# the software as distributed by the Licensor under this license.

# 2. Grant of Rights

# (A) Copyright Grant- Subject to the terms of this license, the Licensor
# grants you a non-transferable, non-exclusive, worldwide, royalty-free
# copyright license to reproduce the software for reference use.

# (B) Patent Grant- Subject to the terms of this license, the Licensor grants
# you a non-transferable, non-exclusive, worldwide, royalty-free patent
# license under licensed patents for reference use.

# 3. Limitations

# (A) No Trademark License- This license does not grant you any rights to use
# the Licensor's name, logo, or trademarks.

# (B) If you begin patent litigation against the Licensor over patents that
# you think may apply to the software (including a cross-claim or counterclaim
# in a lawsuit), your license to the software ends automatically.

# (C) The software is licensed "as-is." You bear the risk of using it. The
# Licensor gives no express warranties, guarantees or conditions. You may have
# additional consumer rights under your local laws which this license cannot
# change. To the extent permitted under your local laws, the Licensor excludes
# the implied warranties of merchantability, fitness for a particular purpose
# and non-infringement.

from flask import Blueprint
from werkzeug.exceptions import default_exceptions, HTTPException

from ... import blueprint_name

app = Blueprint(
    (
        app_name := blueprint_name(__file__)
    ),
    __name__
)


@app.app_errorhandler(HTTPException)
def default_error(error):
    '''
    >>> from flask import abort
    >>> # Cara penggunaan abort ke-1
    >>> abort(400)
    >>> # Cara penggunaan abort ke-2
    >>> abort(400, 'Ini error.')
    >>> # Cara penggunaan abort ke-3
    >>> abort(
    ...  400,
    ...  {
    ...   'paremeter': {'name': 'name harus diisi'},
    ...   'description': 'Ini error.'
    ...  }
    ... )
    '''

    # Google JSON guide:
    # https://stackoverflow.com/questions/12806386/standard-json-api-response-format
    # HTTP status code for “could not fulfill request for *known* reason”:
    # https://stackoverflow.com/questions/33815690/http-status-code-for-could-not-fulfill-request-for-known-reason
    # Flask - How to create custom abort() code?
    # https://stackoverflow.com/questions/12285903/flask-how-to-create-custom-abort-code
    if isinstance(data := error.description, dict):
        parameter = data.get('parameter', {})
        message = data.get('description', None)

        if not message:
            message = default_exceptions[error.code].description
    else:
        parameter = {}
        message = error.description

    return {
        'error_message': message,
        'error_name': error.name,
        'error_parameter': parameter
    }, error.code, '__error__.html.j2'
